
# lib/pdf/writer/fontmetrics.rb
class PDF::Writer::FontMetrics
  class << self
    alias open_org open

    def open(font_name)
      if /(Ryumin|GothicBBB)/ =~ font_name
        metrics = PDF::Writer::FontMetrics.new

        # from PDFJ
        if /Ryumin/ =~ font_name
          metrics.fontbbox = [-170,-331,1024,903]
        else
          metrics.fontbbox = [-174,-268,1001,944]
        end
        metrics.differences = nil
        metrics.c = {}
        metrics.path = ""

        return metrics
      else
        open_org(font_name)
      end
    end
  end
end


# lib/pdf/writer/object/cidsysteminfo.rb
class PDF::Writer::Object::CIDSystemInfo < PDF::Writer::Object
  def initialize(parent)
    super(parent)
  end

  def to_s
    res = "\n#{@oid} 0 obj\n<< /Registry (Adobe) /Ordering (Japan1) /Supplement 2"
    res << ">>\nendobj"
  end
end


# lib/pdf/writer/object/font.rb
class PDF::Writer::Object::Font < PDF::Writer::Object
#  alias :initialize_org :initialize
#
#  def initialize(parent, name, encoding = 'WinAnsiEncoding', subtype = 'Type1')
#    if /(Ryumin|GothicBBB)/ =~ name
#      @cid_sinfo = PDF::Writer::Object::CIDSystemInfo.new(parent)
#    end
#
#    initialize_org(parent, name, encoding, subtype)
#  end

  def initialize(parent, name, encoding = 'WinAnsiEncoding', subtype = 'Type1')
    super(parent)

    @name     = name
    @subtype  = subtype
    @font_id  = @parent.__send__(:generate_font_id)

    if /(Ryumin|GothicBBB)/ =~ name
      @cid_sinfo = PDF::Writer::Object::CIDSystemInfo.new(parent)
    end

    if encoding.kind_of?(PDF::Writer::Object::FontEncoding)
      @encoding           = encoding
    elsif encoding == 'none' or encoding.nil?
      @encoding           = nil
    else
      @encoding           = encoding
    end

    @parent.pages << self

    @firstchar      = nil
    @lastchar       = nil
    @widths         = nil
    @fontdescriptor = nil
  end

  def to_s
    res = "\n#{@oid} 0 obj\n<< /Type /Font\n/Subtype /#{@subtype}\n"
    if /(Ryumin|GothicBBB)/ =~ @name
      res << "/Name /F#{@font_id}\n/BaseFont /#{@name}-#{@encoding}\n"
      res << "/DescendantFonts [<</CIDSystemInfo #{@cid_sinfo.oid} 0 R /BaseFont /#{@name} /Type /Font /DW 1000 /Subtype /CIDFontType0\n"
      res << "/FontDescriptor #{@fontdescriptor} 0 R\n" unless @fontdescriptor.nil?
      res << " /W [231 389 500 631 [500]]>>]\n"
    else
      res << "/Name /F#{@font_id}\n/BaseFont /#{@name}\n"
      res << "/FontDescriptor #{@fontdescriptor} 0 R\n" unless @fontdescriptor.nil?
    end
    if @encoding.kind_of?(PDF::Writer::Object::FontEncoding)
      res << "/Encoding #{@encoding.oid} 0 R\n"
    elsif @encoding
      res << "/Encoding /#{@encoding}\n" if @encoding
    end
    res << "/FirstChar #{@firstchar}\n" unless @firstchar.nil?
    res << "/LastChar #{@lastchar}\n" unless @lastchar.nil?
    res << "/Widths #{@widths} 0 R\n" unless @widths.nil?
    #res << "/FontDescriptor #{@fontdescriptor} 0 R\n" unless @fontdescriptor.nil?
    res << ">>\nendobj"
  end
end


# lib/pdf/writer/object/fontdescriptor.rb
class PDF::Writer::Object::FontDescriptor < PDF::Writer::Object
  def to_s
    res = "\n#{@oid} 0 obj\n<< /Type /FontDescriptor\n"

    if /Ryumin/ =~ @options['FontName']
      res << "/Style <</Panose <010502020300000000000000>>>\n"
    elsif /GothicBBB/ =~ @options['FontName']
      res << "/Style <</Panose <0801020b0500000000000000>>>\n"
    end

    @options.each do |k, v|
      res << "/#{k} #{v}\n" if %w{Ascent CapHeight Descent Flags ItalicAngle StemV AvgWidth Leading MaxWidth MissingWidth StemH XHeight CharSet}.include?(k)
      res << "/#{k} #{v} 0 R\n" if %w{FontFile FontFile2 FontFile3}.include?(k)
      res << "/#{k} [#{v.join(' ')}]\n" if k == "FontBBox"
      res << "/#{k} /#{v}\n" if k == "FontName"
    end
    res << ">>\nendobj"
  end
end


# lib/pdf/writer.rb
class PDF::Writer
  alias :load_font_org :load_font

  def load_font(font, encoding = nil)
    if /(Ryumin|GothicBBB)/ =~ font
      metrics = load_font_metrics(font)

      name  = File.basename(font).gsub(/\.afm$/o, "")

      encoding_diff = nil
      case encoding
      when Hash
        encoding_name = encoding[:encoding]
        encoding_diff = encoding[:differences]
        encoding      = PDF::Writer::Object::FontEncoding.new(self, encoding_name, encoding_diff)
      when NilClass
        encoding_name = encoding = 'WinAnsiEncoding'
      else
        encoding_name = encoding
      end

      metrics.encoding = encoding_name
      wfo  = PDF::Writer::Object::Font.new(self, name, encoding, 'Type0')
      fdsc = PDF::Writer::Object::FontDescriptor.new(self)
      if /Ryumin/ =~ font
        fdopt = {
          'Ascent'    => 723,
          'CapHeight' => 709,
          'Descent'   => -241,
          'Flags'     => 6,
          'StemV'     => 69,
          'XHeight'   => 450,
          'FontBBox'  => [-170,-331,1024,903],
        }
      else # GothicBBB
        fdopt = {
          'Ascent'    => 752,
          'CapHeight' => 737,
          'Descent'   => -271,
          'Flags'     => 4,
          'StemV'     => 99,
          'XHeight'   => 553,
          'FontBBox'  => [-174,-268,1001,944],
        }
      end
      fdopt['FontName']    = font
      fdopt['ItalicAngle'] = 0
      fdsc.options = fdopt
      wfo.__send__("fontdescriptor=".intern, fdsc.oid)
      return metrics
    else
      load_font_org(font, encoding)
    end
  end

  def add_text(x, y, text, size = nil, angle = 0, word_space_adjust = 0)
    if text.kind_of?(Numeric) and size.kind_of?(String)
      text, size = size, text
      warn PDF::Writer::Lang[:add_text_parameters_reversed] % caller[0]
    end

    if size.nil? or size <= 0
      size = @font_size
    end

    select_font("Helvetica") if @fonts.empty?

    text = text.to_s

      # If there are any open callbacks, then they should be called, to show
      # the start of the line
    @callbacks.reverse_each do |ii|
      info = ii.dup
      info[:x]      = x
      info[:y]      = y
      info[:angle]  = angle
      info[:status] = :start_line

      info[:tag][self, info]
    end
    if angle == 0
      add_content("\nBT %.3f %.3f Td" % [x, y])
    else
      rad = PDF::Math.deg2rad(angle)
      tt = "\nBT %.3f %.3f %.3f %.3f %.3f %.3f Tm"
      tt = tt % [ Math.cos(rad), Math.sin(rad), -Math.sin(rad), Math.cos(rad), x, y ]
      add_content(tt)
    end

    if (word_space_adjust != 0) or not ((@word_space_adjust.nil?) and (@word_space_adjust != word_space_adjust))
      @word_space_adjust = word_space_adjust
      add_content(" %.3f Tw" % word_space_adjust)
    end

    pos = -1
    start = 0
    loop do
      pos += 1
      break if pos == text.size
      font_change = true
      tag_size, text, font_change = quick_text_tags(text, pos, font_change)

      if tag_size != 0
        if pos > start
          part = text[start, pos - start]
          tt = " /F#{find_font(@current_font).font_id}"
          tt << " %.1f Tf %d Tr" % [ size, @current_text_render_style ]
          tt << " (#{PDF::Writer.escape(part)}) Tj"
          add_content(tt)
        end

        if font_change
          current_font!
        else
          add_content(" ET")
          xp = x
          yp = y
          tag_size, text, font_change, xp, yp = text_tags(text, pos, font_change, true, xp, yp, size, angle, word_space_adjust)

            # Restart the text object
          if angle.zero?
            add_content("\nBT %.3f %.3f Td" % [xp, yp])
          else
            rad = PDF::Math.deg2rad(angle)
            tt = "\nBT %.3f %.3f %.3f %.3f %.3f %.3f Tm"
            tt = tt % [ Math.cos(rad), Math.sin(rad), -Math.sin(rad), Math.cos(rad), xp, yp ]
            add_content(tt)
          end

          if (word_space_adjust != 0) or (word_space_adjust != @word_space_adjust)
            @word_space_adjust = word_space_adjust
            add_content(" %.3f Tw" % [word_space_adjust])
          end
        end

        pos += tag_size - 1
        start = pos + 1
      end
    end

    if start < text.size
      part = text[start..-1]

      tt = " /F#{find_font(@current_font).font_id}"
      tt << " %.1f Tf %d Tr" % [ size, @current_text_render_style ]
      encoding = @fonts[@current_font].encoding
      if /(Ryumin|GothicBBB)/ =~ @current_font
        hex = ja_str_encode(part, encoding)
        tt << " [#{hex}] TJ"
      else
        tt << " (#{PDF::Writer.escape(part)}) Tj"
      end
      add_content(tt)
    end
    add_content(" ET")

      # XXX: Experimental fix.
    @callbacks.reverse_each do |ii|
      info = ii.dup
      info[:x]      = x
      info[:y]      = y
      info[:angle]  = angle
      info[:status] = :end_line
      info[:tag][self, info]
    end
  end

  def char_width(font, char, ja_position = nil)
    char = char[0] unless @fonts[font].c[char]

    if /(Ryumin|GothicBBB)/ =~ font
      if ja_position == :FIRST ||
         (ja_position == :MAYBE_THIRD && !is_ja_char_pos(char, :THIRD, :UTF8))
        if char < 0x81 # hankaku # ok? at UTF-8
          return @fonts[font].fontbbox[2].to_f / 2
        else # zenkaku
          return @fonts[font].fontbbox[2].to_f
        end
      else
        return 0
      end
    end

    if @fonts[font].differences and @fonts[font].c[char].nil?
      name = @fonts[font].differences[char] || 'M'
      width = @fonts[font].c[name]['WX'] if @fonts[font].c[name]['WX']
    elsif @fonts[font].c[char]
      width = @fonts[font].c[char]['WX']
    else
      width = @fonts[font].c['M']['WX']
    end
    width
  end

  def text_line_width(text, size = nil)
    if text.kind_of?(Numeric) and size.kind_of?(String)
      text, size = size, text
      warn PDF::Writer::Lang[:text_width_parameters_reversed] % caller[0]
    end

    if size.nil? or size <= 0
      size = @font_size
    end

      # This function should not change any of the settings, though it will
      # need to track any tag which change during calculation, so copy them
      # at the start and put them back at the end.
    t_CTS = @current_text_state.dup

    select_font("Helvetica") if @fonts.empty?
      # converts a number or a float to a string so it can get the width
    tt = text.to_s
      # hmm, this is where it all starts to get tricky - use the font
      # information to calculate the width of each character, add them up
      # and convert to user units
    width = 0
    font = @current_font
    encoding = @fonts[font].encoding
    ja_position = :FIRST

    pos = -1
    loop do
      pos += 1
      break if pos == tt.size
      font_change = true
      tag_size, text, font_change = quick_text_tags(text, pos, font_change)
      if tag_size != 0
        if font_change
          current_font!
          font = @current_font
        end
        pos += tag_size - 1
      else
        if "&lt;" == tt[pos, 4]
          width += char_width(font, '<')
          pos += 3
        elsif "&gt;" == tt[pos, 4]
          width += char_width(font, '>')
          pos += 3
        elsif "&amp;" == tt[pos, 5]
          width += char_width(font, '&')
          pos += 4
        else
          width += char_width(font, tt[pos, 1], ja_position)
          ja_position = ja_char_position(tt[pos], ja_position, encoding)
        end
      end
    end

    @current_text_state = t_CTS.dup
    current_font!

    (width * size / 1000.0)
  end

  def add_text_wrap(x, y, width, text, size = nil, justification = :left, angle = 0, test = false)
    if text.kind_of?(Numeric) and size.kind_of?(String)
      text, size = size, text
      warn PDF::Writer::Lang[:add_textw_parameters_reversed] % caller[0]
    end

    if size.nil? or size <= 0
      size = @font_size
    end

      # Need to store the initial text state, as this will change during the
      # width calculation, but will need to be re-set before printing, so
      # that the chars work out right
    t_CTS = @current_text_state.dup

    select_font("Helvetica") if @fonts.empty?
    return "" if width <= 0

    w = brk = brkw = 0
    font = @current_font
    tw = width / size.to_f * 1000
    encoding = @fonts[font].encoding
    ja_position = :FIRST

    pos = -1
    loop do
      pos += 1
      break if pos == text.size
      font_change = true
      tag_size, text, font_change = quick_text_tags(text, pos, font_change)
      if tag_size != 0
        if font_change
          current_font!
          font = @current_font
        end
        pos += (tag_size - 1)
      else
        w += char_width(font, text[pos, 1], ja_position)
        ja_position = ja_char_position(text[pos], ja_position, encoding)

        if w > tw # We need to truncate this line
          if brk > 0 # There is somewhere to break the line.
            if text[brk] == " "
              tmp = text[0, brk]
            else
              tmp = text[0, brk + 1]
            end
            x, adjust = adjust_wrapped_text(tmp, brkw, width, x, justification)

              # Reset the text state
            @current_text_state = t_CTS.dup
            current_font!
            add_text(x, y, tmp, size, angle, adjust) unless test
            return text[brk + 1..-1]
          else # just break before the current character
            # adjust for ja
            if /RKSJ/ =~ encoding
              pos -= 1 if !text.empty? && is_ja_char_pos(text[pos-1], :FIRST, :SJIS)
            elsif /EUC/ =~ encoding
              #pos -= 1 if !text.empty? && is_ja_char_pos(text[pos-1], :FIRST, :EUC)
              if /([\xa1-\xfe][\xa1-\xfe])+[\xa1-\xfe]$/e =~ text[0, pos] ||
                 /[\x20-\x7e][\xa1-\xfe]$/e =~ text[0, pos]
                pos -= 1
              end              
            else # UTF-8
              #pos -= 1 if /[\xc0-\xdf]|[\xe0-\xef]$/u =~ text[0, pos]
              if text.length > pos &&
                 is_ja_char_pos(text[pos], :SECOND, :UTF8) # :SECOND or :THIRD
                if is_ja_char_pos(text[pos-1], :SECOND, :UTF8)
                  # [\xe0-\xef][\x80-\xbf]  [\x80-\xbf]
                  pos -= 2
                else
                  # [\xc0-\xdf]  [\x80-\xbf]
                  # [\xe0-\xef]  [\x80-\xbf][\x80-\xbf]
                  pos -= 1
                end
              end
            end

            tmp = text[0, pos]
#           tmpw = (w - char_width(font, text[pos, 1])) * size / 1000.0
            x, adjust = adjust_wrapped_text(tmp, brkw, width, x, justification)

              # Reset the text state
            @current_text_state = t_CTS.dup
            current_font!
            add_text(x, y, tmp, size, angle, adjust) unless test
            return text[pos..-1]
          end
        end

        if text[pos] == ?-
          brk = pos
          brkw = w * size / 1000.0
        end

        if text[pos, 1] == " "
          brk = pos
          ctmp = text[pos]
          ctmp = @fonts[font].differences[ctmp] unless @fonts[font].differences.nil?
          z = @fonts[font].c[tmp].nil? ? 0 : @fonts[font].c[tmp]['WX']
          brkw = (w - z) * size / 1000.0
        end
      end
    end

      # There was no need to break this line.
    justification = :left if justification == :full
    tmpw = (w * size) / 1000.0
    x, adjust = adjust_wrapped_text(text, tmpw, width, x, justification)
      # reset the text state
    @current_text_state = t_CTS.dup
    current_font!
    add_text(x, y, text, size, angle, adjust) unless test
    return ""
  end


  def is_ja_char_pos(c, pos, encoding)
    case encoding
    when :SJIS
      case pos
      when :FIRST
        (c >= 0x81 && c <= 0x9f) || (c >= 0xe0 && c <= 0xef)
      when :SECOND
        (c >= 0x40 && c <= 0x7e) || (c >= 0x80 && c <= 0xfc)
      else
      end
    when :EUC
      case pos
      when :FIRST
        c >= 0xa1 && c <= 0xfe
      when :SECOND
        c >= 0xa1 && c <= 0xfe
      else
      end
    when :UTF8
      case pos
      when :FIRST
        (c >= 0xc0 && c <= 0xdf) || (c >= 0xe0 && c <= 0xef)
      when :SECOND, :THIRD
        c >= 0x80 && c <= 0xbf
      else
      end
   else
      raise "Unkowwn encoding!"
    end
  end

  def toutf16(str)
    begin
      require 'uconv'
      Uconv.u16swap(Uconv.u8tou16(str))
    rescue LoadError
      require 'kconv.rb'
      Kconv.kconv(str, NKF::UTF16, NKF::UTF8)
      #str.toutf16
    end
  end

  def ja_char_position(c, current_position, encoding)
    #p [current_position, sprintf("0x%02x", c)]

    if /RKSJ/ =~ encoding
      if current_position == :FIRST && is_ja_char_pos(c, :FIRST, :SJIS)
        current_position = :SECOND
      elsif current_position == :SECOND
        current_position = :FIRST # no error check
      end
    elsif /EUC/ =~ encoding
      if current_position == :FIRST && is_ja_char_pos(c, :FIRST, :EUC)
        current_position = :SECOND
      elsif current_position == :SECOND
        current_position = :FIRST # no error check
      end
    else # UTF-8
      if current_position == :FIRST
        if is_ja_char_pos(c, :FIRST, :UTF8)
          current_position = :SECOND
        end
      elsif current_position == :SECOND
        current_position = :MAYBE_THIRD
      elsif current_position == :MAYBE_THIRD
        if is_ja_char_pos(c, :THIRD, :UTF8)
          current_position = :FIRST
        else
          current_position = :SECOND
        end
      end
    end
    current_position
  end

  def ja_str_encode(str, encoding)
    hex = ""; pre_c = :START; now_c = :START

    if /RKSJ/ =~ encoding
      a = PDF::Writer.escape(str.chomp).split(//sm)
    elsif /EUC/ =~ encoding
      a = PDF::Writer.escape(str.chomp).split(//em)
    else
      a = PDF::Writer.escape(str.chomp).split(//um)
    end

    if /(RKSJ|EUC)/ =~ encoding
      enc = :NON_UNICODE
    else
      enc = :UNICODE
    end

    a.each do |c|
      c = toutf16(c) if enc == :UNICODE

      if enc == :UNICODE || c.size > 1
        now_c = :JA
        hex << ")" if pre_c == :ASCII
        hex << "<" if pre_c != :JA
        c.each_byte {|x| hex << (sprintf "%02x", x)}
      else
        now_c = :ASCII
        hex << ">" if pre_c == :JA
        hex << "(" if pre_c != :ASCII
        hex << c
      end
      pre_c = now_c
    end
    hex << ">" if pre_c == :JA
    hex << ")" if pre_c == :ASCII
    hex
  end
end
